# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/20_extracting-image-data.ipynb.

# %% auto 0
__all__ = ['extract_jpg']

# %% ../notebooks/20_extracting-image-data.ipynb 16
from . import file_to_bytes, get_blocks, get_blocktypes 
import numpy as np 
import io 
import re
import matplotlib.pyplot as plt 
from PIL import Image

# %% ../notebooks/20_extracting-image-data.ipynb 17
def extract_jpg(pdz_file, BLOCKTYPE=137, save_file=False):
    '''Extract and save jpg images from `pdz_file`.
    
    Returns a list of jpg images where ims = [ im0, im1, ... ].
    '''

    # parse into blocks
    pdz_bytes = file_to_bytes(pdz_file)
    block_list = get_blocks(pdz_bytes, verbose=False)

    # read block 137 (if present)

    blocktypes_list = get_blocktypes(block_list)

    if BLOCKTYPE not in blocktypes_list:

        print(f'Could not find jpg image data in: {pdz_file}')

    else:
        jpg_i = blocktypes_list.index(BLOCKTYPE)
        jpg_dict = block_list[jpg_i]

        jpg_sandwich = jpg_dict['bytes'].tobytes()

        # code below thanks to Lars Maxfield
        # Repeatedly search for jpgs by consuming jpg_sandwich

        ims = []
        while True:
            match_jpg_start = re.search(b'\xff\xd8', jpg_sandwich)
            match_jpg_end = re.search(b'\xff\xd9', jpg_sandwich)

            if not match_jpg_start or not match_jpg_end:
                
                break

            jpg_start = match_jpg_start.span()[0]
            jpg_end = match_jpg_end.span()[1]
            jpg = jpg_sandwich[jpg_start:jpg_end]

            ims.append(np.array(Image.open(io.BytesIO(jpg))))

            jpg_sandwich = jpg_sandwich[jpg_end:]

        if save_file is True:
            for i, im in enumerate(ims):
                jpg_file = re.sub('\.pdz$', f'-{i}.jpg', pdz_file)
                
                print(f"Saving image file: '{jpg_file}'")
                plt.imsave(jpg_file, im)

        return ims

    return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/30_parsing-bytes.ipynb.

# %% auto 0
__all__ = ['PDZ_25_STRUCTURE_DICT', 'extract_spectra', 'multiparse', 'prefix', 'file_to_bytes', 'parse', 'read_strings',
           'skip_bytes', 'read_table', 'read_counts', 'get_block_at', 'get_blocks', 'get_blocktypes']

# %% ../notebooks/30_parsing-bytes.ipynb 47
import struct 
import numpy as np 
import os 
import re
import matplotlib.pyplot as plt
import pandas as pd 
from IPython.display import display

# %% ../notebooks/30_parsing-bytes.ipynb 48
PDZ_25_STRUCTURE_DICT = {
    25:  {'xformat': 'hi-10X-i', 
          'param_keys': ['pdz_type', 'block_size', 'FileFormatString?', '??']}, 
    1:   {'xformat': 'hi-2S-6s-2S-h-S-T', 
          'param_keys': ['block_type', 'block_size', '??', 'SerialString', '??', '??', '??', '??', '??', '??', 
                         '??', '??', '??', '??', '??', '??', '??']}, 
    2:   {'xformat': 'hi3i8f-*X', 
          'param_keys': ['block_type', 'block_size', '??', 'RawCounts', 'ValidCounts', '??', '??', 
                         '??', 'ActiveTimeInSeconds', 'DeadTimeInSeconds', 'ResetTimeInSeconds', 
                         'LiveTimeInSeconds', 'TotalElapsedTimeInSeconds', '??']}, 
    3:   {'xformat': 'hi-3i9f7hfhfhfhf8hfhi-S-h-Z', 
          'param_keys': ['block_type', 'block_size', '??', 'RawCounts', 'ValidCounts', 
                         '??', '??', '??', 'ActiveTimeInSeconds', 'DeadTimeInSeconds', 
                         'ResetTimeInSeconds', 'LiveTimeInSeconds', 'XrayVoltageInkV', 'XrayFilamentCurrentInMicroAmps', 
                         'Filter1ElementAtomicNum', 'Filter1Thickness', 'Filter2ElementAtomicNum', 'Filter2Thickness', 
                         'Filter3ElementAtomicNum', 'Filter3Thickness', '??', 'DetectorTempInC', '??', 
                         '??', '??', 'eVPerChannel', '??', 'eVStart', 
                         'Year', 'Month', 'AM/PM code?', 'Day', 'Hour', 'Minutes', 'Seconds', 
                         '??', 'NosePressureInMilliBars', 'NumberOfChannels', 'NoseTemperatureInC', 
                         'TubeSpec?', '??', '2048 counts']}}


def extract_spectra(pdz_file, to_csv=True, verbose=True): 
    '''Directly extract spectral data from `pdz_file`. 
    
    Robust extraction of spectrum counts and energy calibration. Other meta data is ignored. '''

    pdz_bytes = file_to_bytes(pdz_file)
    block_list = get_blocks(pdz_bytes, verbose=False) 
    
    # select type 3 blocks
    b3_list = [b for b in block_list if b['block_type'] == 3] 
    n_spectra = len(b3_list) 

    # parsing spectrum parameters 
    #from first block to compute channel energies (keV) 
    # (assuming that these are similar for all spectra in the pdz file) 
    
    arr = b3_list[0]['bytes'] # only using first spectrum!  
    spectrum_params = multiparse('hi-3i9f7hfhfhfhf8hfhi-S-h', arr, verbose=False)[0]  
    tube_keV = spectrum_params[12] # FYI 
    delta_keV = spectrum_params[25] / 1000
    start_keV = spectrum_params[27] / 1000 
    n_channels = spectrum_params[37] 
    if n_channels != 2048: 
        print(f'Found unexpected number of channels in pdz metadata: {n_channels}')
        
    stop_keV = start_keV + delta_keV * (n_channels -1)
    x_keV = np.linspace(start_keV, stop_keV, num=n_channels) 

    # initialize array 
    
    spectra_df = pd.DataFrame(index=x_keV)

    # TODO: Test if computed channel energies `x_keV` are reasonable, 
    # Parsing of spectrum parameters might fail 
    # if Bruker messes with file format.   
    # otherwise fall back on 0-40 keV range  

    spectrum_list = [] 

    for i, b3 in enumerate(b3_list):  
        arr = b3['bytes'] 
        counts = np.array(parse(f'{n_channels}i', arr[-4*n_channels:], verbose=False)[0])  

        spectra_df[f'spectrum #{i+1}'] = counts

    if to_csv == True: 
        csv_file = f'{pdz_file}.csv' 
        print(f'Saving spectral data to: {csv_file}')
        spectra_df.to_csv(csv_file, float_format='%10.5f')

    return spectra_df
           

def multiparse(xformat, arr, param_keys=None, verbose=True): 
    '''Parse segments in extendend format string `xformat` e.g. '<i5f-2S-T-3S-S-f' '''

    
    parts = re.split('-', xformat) 

    result_list = []
    for p in parts: 
        if 'S' in p:
            result, arr = read_strings(p, arr, verbose=False) 
        elif p == 'T': 
            result, arr = read_table(p, arr, verbose=False) 
        elif 'X' in p: 
            result, arr = skip_bytes(p, arr, verbose=False) 
        elif 'Z' in p: 
            result, arr = read_counts(p, arr, verbose=False)
            result = [result]
        elif 'z' in p: 
            result, arr = read_counts(p, arr, n_channels=1028, verbose=False)
            result = [result]    
        
        else: 
            result, arr = parse(p, arr, verbose=False) 
            
        result_list.extend(result)   

    if verbose: 
        if param_keys == None: 
            result_df = pd.DataFrame({'values': result_list})
        else: 
            result_df = pd.DataFrame({'values': result_list, 'param_keys': param_keys})      
        display(result_df)

    return result_list, arr   

def prefix(format): 
    '''Prefix little endian byte order (<) to struct type format string if missing. '''

    if not format.startswith('<'): 
        format = f'<{format}'
    
    return format    

def file_to_bytes(pdz_file): 
    '''Read all bytes from filepath `pdz_file` into a byte array. 
    
    Returns: `pdz_arr` (numpy array of bytes)
    '''

    with open(pdz_file, 'rb') as fh: 
        blob = fh.read() 
        
    pdz_arr = np.array([v[0] for v in struct.iter_unpack('c', blob)])
    #pdz_arr = bytearray(blob)

    return pdz_arr 
    

def parse(format, arr, verbose=True): 
    '''Parse first bytes from bytes array `arr` into human readable text according to `format` string. 
    
    See struct library for format string specification. For example, '<ff' would result 
    the first 8 bytes to be converted into two Little-Endian floats. 
    
    Returns: `parsed` list and remaining bytes array of `tail_arr` unprocessed `values`.'''

    format = prefix(format) 
    size = struct.calcsize(format)
    buffer = arr[0:size]
    tail_arr = arr[size:]

    parsed = list(struct.unpack(format, buffer)) 

    if verbose: 
        print(parsed)
    
    return parsed, tail_arr

def read_strings(xformat, arr, verbose=True): 
    '''Parse `n` variable length character strings preceded by a length integer. 
     
    '''
    # get multiplier  
    if xformat == 'S': 
        n = 1
    else: 
        n = int(re.sub('(^\d+)S', '\g<1>', xformat))

    # parse strings 
    string_list = [] 
    while n > 0: 
        [length], arr = parse('<i', arr, verbose=False) # read length 
        n_bytes = 2 * length 

        # do some testing 
        assert (n_bytes > 1) and (type(n_bytes) is int), f'{n_bytes} is invalid string length' 
        
        char_list, arr = parse(f'<{n_bytes}c', arr, verbose=False) 
        string = b''.join(char_list).decode(encoding='utf-16') 
        string_list.append(string) 
        n = n -1 
        
    if verbose: 
        print(string_list) 

    return string_list, arr


def skip_bytes(xformat, arr, verbose=True): 
    '''Skip a number of bytes as specified in `xformat` string. 
    
    If multiplier is `*` then skip all. 
    '''

    # get multiplier  
    if xformat == 'X': 
        n_bytes = 1
    elif xformat == '*X': 
        n_bytes = len(arr)    
    else: 
        n_bytes = int(re.sub('(^\d+)X', '\g<1>', xformat))

    skipped = [b''.join(arr[0:n_bytes])] 

    arr = arr[n_bytes:] 

    if verbose: 
        print(skipped)

    return skipped, arr 
    

def read_table(xformat, arr, verbose=True): 
    '''Extract numbered table'''

    assert xformat == 'T', 'Incorrect format string'

    [table_length], arr = parse('<i', arr, verbose=False) 

    table = []
    for i in range(table_length): 
        [num], arr = parse('<h', arr, verbose=False)  
        [string], arr = read_strings('S', arr, verbose=False)
        table.append([f'#{num}', string]) 
        
    if verbose: 
        print(table)

    return table, arr


def read_counts(xformat, arr, n_channels=2048, verbose=True): 
    '''Extract counts. '''

    assert xformat == 'Z'  or xformat == 'z', 'Incorrect format string'

    format = f'<{n_channels}i'

    counts, arr = parse(format, arr, verbose=False) 
    counts = np.array(counts)
        
    if verbose: 
        print(counts)

    return counts, arr
    

def get_block_at(pdz_arr, start): 
    '''Read first data block from bytes array `pdz_arr` from index position `start`. 

    Assumes that first 4 bytes are (block type and size)
    
    Returns: `block_dict`, `block` 
    '''

    file_size = len(pdz_arr)
    
    [block_type, block_size], arr = parse('hi', pdz_arr[start:], verbose=False)
    
    stop = start + block_size + 6 # four bytes extra due to `dtype` and `size` shorts plus two empty pad bytes? 

    # read block bytes
    arr = pdz_arr[start:stop] 
    
    block_dict = {'block_type': block_type, 'block_size': block_size, 'start': start, 'stop': stop, 
                  'file_size': file_size, 'bytes': arr}

    return block_dict


def get_blocks(pdz_bytes, verbose=True): 
    '''Parse `pdz_byte_array` into consequtive blocks. ''' 

    if verbose: 
        print('Detecting block sequence...')
    
    start = 0 
    total_size = len(pdz_bytes) 

    block_list = [] 
    while start < total_size: 
        block_dict = get_block_at(pdz_bytes, start)
        start = block_dict['stop']
        block_list.append(block_dict)

    else:
        if start == total_size: 
            if verbose: 
                print('Ok!')
        else: 
            print('Error while reading last block! ')
            print(f'Stop index: {start} does not match total file size: {total_size}')
        
    return block_list


def get_blocktypes(block_list): 
    '''Extract `block_types` list from `block_list`. '''
    
    block_types = []

    for block_dict in block_list: 
        t = block_dict['block_type']
        block_types.append(t)

    return block_types




# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/40_pdz-archeology.ipynb.

# %% auto 0
__all__ = ['PDZ11_STRUCT_DICT', 'check_pdz_type']

# %% ../notebooks/40_pdz-archeology.ipynb 16
from . import file_to_bytes, get_blocks, multiparse, parse 
import struct

# %% ../notebooks/40_pdz-archeology.ipynb 17
PDZ11_STRUCT_DICT = {
    'pdz11_2048_channels' : {'xformat': '2X-4X-h-34X-2d-86X-2i-10X-2f-188X-Z-*X', 
                          'param_keys': ['pdz-version', '??', 'NumberOfChannels', '??', 'LiveTimeInSeconds', 
                                         'eVPerChannel', '??', 'RawCounts', 'ValidCounts', '??',  'XrayVoltageInkV', 
                                         'XrayFilamentCurrentInMicroAmps', '??', 'PhotonCounts (2048 channels)', '??']}, 
    'pdz11_1024_channels' : {'xformat': '2X-4X-h-34X-2d-86X-2i-10X-2f-24X-z-*X', 
                          'param_keys': ['pdz-version', '??', 'NumberOfChannels', '??', 'LiveTimeInSeconds', 
                                         'eVPerChannel', '??', 'RawCounts', 'ValidCounts', '??',  'XrayVoltageInkV', 
                                         'XrayFilamentCurrentInMicroAmps', '??', 'PhotonCounts (1024 channels)', '??']} 
}


def check_pdz_type(pdz_file, verbose=True): 
    '''Read first two bytes and for legacy pdz files number of detector channels to check pdz file type.'''

    pdz_bytes = file_to_bytes(pdz_file) 
    file_size = len(pdz_bytes)
    first_two_bytes = struct.unpack('<h', pdz_bytes[0:2])[0] 

    if first_two_bytes == 25: 
        pdz_type = 'pdz25' 
    elif first_two_bytes == 257: 
        n_channels = struct.unpack('<h', pdz_bytes[6:8])[0] 
        if n_channels == 1024:
            pdz_type = 'pdz11_1024_channels'
        elif n_channels == 2048:
            pdz_type = 'pdz11_2048_channels'  
        else: 
            pdz_type = f'pdz11_with_unexpected_number_of_{n_channels}_channels'
    else:
        pdz_type = f'pdz_type_unknown:{first_two_bytes}'
                
    return pdz_type



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/read_pdz',
                'doc_host': 'https://fligt.github.io',
                'git_url': 'https://github.com/fligt/read_pdz',
                'lib_path': 'read_pdz'},
  'syms': { 'read_pdz.jpg_extractor': { 'read_pdz.jpg_extractor.extract_jpg': ( 'extracting-image-data.html#extract_jpg',
                                                                                'read_pdz/jpg_extractor.py')},
            'read_pdz.legacy': {'read_pdz.legacy.check_pdz_type': ('pdz-archeology.html#check_pdz_type', 'read_pdz/legacy.py')},
            'read_pdz.reader': { 'read_pdz.reader.extract_spectra': ('parsing-bytes.html#extract_spectra', 'read_pdz/reader.py'),
                                 'read_pdz.reader.file_to_bytes': ('parsing-bytes.html#file_to_bytes', 'read_pdz/reader.py'),
                                 'read_pdz.reader.get_block_at': ('parsing-bytes.html#get_block_at', 'read_pdz/reader.py'),
                                 'read_pdz.reader.get_blocks': ('parsing-bytes.html#get_blocks', 'read_pdz/reader.py'),
                                 'read_pdz.reader.get_blocktypes': ('parsing-bytes.html#get_blocktypes', 'read_pdz/reader.py'),
                                 'read_pdz.reader.multiparse': ('parsing-bytes.html#multiparse', 'read_pdz/reader.py'),
                                 'read_pdz.reader.parse': ('parsing-bytes.html#parse', 'read_pdz/reader.py'),
                                 'read_pdz.reader.prefix': ('parsing-bytes.html#prefix', 'read_pdz/reader.py'),
                                 'read_pdz.reader.read_counts': ('parsing-bytes.html#read_counts', 'read_pdz/reader.py'),
                                 'read_pdz.reader.read_strings': ('parsing-bytes.html#read_strings', 'read_pdz/reader.py'),
                                 'read_pdz.reader.read_table': ('parsing-bytes.html#read_table', 'read_pdz/reader.py'),
                                 'read_pdz.reader.skip_bytes': ('parsing-bytes.html#skip_bytes', 'read_pdz/reader.py')}}}
